from flask import Blueprint, abort, current_app, send_from_directory

blueprint = Blueprint('static', __name__)

@blueprint.route('/<path:path>')
def get_static(path):
    return send_from_directory(current_app.static_folder, path)



import yaml
import os
import pkg_resources
import pathlib

def default_config():
    getpath = lambda filename: str(pathlib.Path(pkg_resources.resource_filename(__name__, filename)).resolve())

    return {
        "flask": {
            "template_folder": getpath("templates"),
            "static_folder": getpath("static")
        },
        "jinja2": {
            "trim_blocks": True,
            "lstrip_blocks": True
        }
    }

def load_config(path = None):
    if path is None:
        path = os.getenv('PWMOOSE_CONFIG_PATH')

    if path is None:
        path = pkg_resources.resource_filename(__name__, "config.yml")

    config = default_config()
    with open(path, 'r') as fin:
        config.update(yaml.safe_load(fin.read()))

    return config



from flask import render_template

class CatchableException(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message

    def __repr__(self):
        return self.message

def exception_to_error(app, template):
    def exception_to_error_2(func):
        def exception_to_error_3(*args, **kwargs):
            try:
                ret = func(*args, **kwargs)
            except CatchableException as e:
                return render_template(template, error_message=repr(e))
            else:
                return ret
        exception_to_error_3.__name__ = func.__name__
        return exception_to_error_3
    return exception_to_error_2



import re

from flask import Blueprint, abort, current_app, render_template, redirect

class NormalizePackageName:
    re_normalize = re.compile(r'[-_\.]+')

    @classmethod
    def normalize(cls, package_name):
        return cls.re_normalize.sub("-", package_name).lower()

blueprint = Blueprint('package_file_list', __name__)

@blueprint.route('/<string:package_name>')
def PackageFileListRedirect(package_name):
    return redirect(f'/{package_name}/', code=302)

@blueprint.route('/<string:package_name>/')
def PackageFileList(package_name):
    if len(package_name) > 1024:
        abort(400, "Package names must be less than 1024 characters in length")

    if not package_name.isascii():
        abort(400, "Package name must be made up of only ascii characters")

    normalized_package_name = NormalizePackageName.normalize(package_name)
    if normalized_package_name != package_name:
        return redirect(f'/{normalized_package_name}', code=302)

    results = None
    for repo in current_app.repositories:
        list_result = repo.list_package_files(package_name)
        if list_result.ok:
            results = list_result.get_results()
            break

    if results is None:
        abort(404, "Package not found")

    return render_template("package_file_list.html", list_results=results, package_name=package_name)

@blueprint.route('')
def PackageListRedirect():
    return redirect(f'', code=302)

@blueprint.route('/')
def PackageList():
    all_packages = set()
    for repo in current_app.repositories:
        current_app.logger.debug(f"Requesting list of packages from {repo.baseurl}")
        list_result = repo.list_packages()
        current_app.logger.debug("Updating package list")
        if list_result.ok:
            all_packages.update([pkg.name for pkg in list_result.get_results()])
    current_app.logger.debug("Collected package list")

    return render_template("package_list.html", list_results=list(sorted(all_packages)))



from urllib.parse import ParseResult, urlunparse, urlencode, quote
from contextlib import contextmanager
import base64
import datetime
import functools

from flask import current_app, abort
from flask.json import JSONEncoder

from .exception import CatchableException

class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, datetime.datetime):
            return obj.isoformat()
        return JSONEncoder.default(self, obj)

def now():
    return datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)



from flask import abort
from .ABCRepository import ABCRepository, PackageFileListResult, PackageListResult
import requests
import urllib.parse
import pathlib

class FilesystemRepository(ABCRepository):
    def __init__(self, *, path, **kwargs):
        super().__init__(**kwargs)
        self.path = pathlib.Path(path)

    def list_all_packages(self):
        for child in self.path.iterdir():
            if child.is_dir():
                link = f'{child.name}/'
                result.add_result(name = child.name, link=link)

        return result

    def list_all_package_files(self, package_name):
        for child in (self.path / package_name).iterdir():
            if child.is_file():
                link = f'{package_name}/{child.name}'
                result.add_result(name = child.name, link=link)

        return result

    def get_package_file(self, package_name, file_name):
        file_path = self.path / package_name / file_name

        if not file_path.exists():
            abort(404)

        with file_path.open('rb') as fin:
            yield fin




import logging
import logging.config
import inspect

from flask import Flask

from .exception import CatchableException
#from .database import connect
from . import config
from . import blueprints
from . import repositories

from .util import CustomJSONEncoder

def create_app(name = None, cfg = None):
    if cfg is None:
        cfg = config.load_config()

    if name is None:
        name = __name__

    if 'logging' in cfg:
        logging.config.dictConfig(cfg['logging'])

    app = Flask(name, **cfg['flask'])
    app.config['RESTFUL_JSON'] = {
        'cls': CustomJSONEncoder
    }
    app.site_config = cfg

    app.repositories = []
    repo_config = cfg['repository']
    repo_class = repo_config['class']
    if repo_class.startswith("_"):
        raise ValueError("Repository class names cannot start with '_'")
    if not hasattr(repositories, repo_class):
        raise ValueError(f"Invalid repository class: {repo_class}")

    Repo = getattr(repositories, repo_class)

    if inspect.isabstract(Repo):
        raise ValueError(f"Cannot instantiate abstract repository class: {Repo}")

    repo_spec = repo_config.get('spec', {})
    print(repo_spec, repo_config)
    app.repositories.append(Repo(**repo_spec))

    app.jinja_options.update(cfg['jinja2'])
    #app.dbengine, app.dbsession_factory = connect(cfg['database'])
    #app.session = None

    app.register_blueprint(blueprints.Static.blueprint, url_prefix='/static')
    app.register_blueprint(blueprints.Index.blueprint, url_prefix='/simple')

    @app.before_request
    def log_request_info():
        from flask import request
        app.logger.debug(f'Got {request.method} request with mimetype: {request.mimetype} for {request.full_path}')
        headers = []
        for name, value in request.headers:
            headers.append((name, value))
        app.logger.debug(f'Header: {headers}')

#    app.json_encoder = CustomJSONEncoder

#    @app.errorhandler(CatchableException)
#    def exception_to_error_handler(e):
#        return render_template("pages/exception.html", error_message=repr(e))

    return app



import abc
import collections

class ListResult:
    def __init__(self, ok):
        self.results = []
        self.ok = ok

    @abc.abstractmethod
    def add_result(self, **kwargs):
        pass

    def get_results(self):
        return self.results

class PackageListResult(ListResult):
    PackageResult = collections.namedtuple(
        'PackageResult', ['name', 'link']
    )

    def add_result(self, **kwargs):
        self.results.append(self.PackageResult(**kwargs))

class PackageFileListResult(ListResult):
    PackageFileResult = collections.namedtuple(
        'PackageFileResult', ['name', 'link', 'requires_python']
    )

    def add_result(self, **kwargs):
        self.results.append(self.PackageFileResult(**kwargs))

class ABCRepository(abc.ABC):
    def __init__(
            self, *, exclude_packages = None, include_packages = None,
            include_package_files = None, exclude_package_files = None,
            stop_on_package_match = True
    ):
        def coalesce(val, default):
            if val is not None:
                return val
            else:
                return default

        self.stop_on_package_match = stop_on_package_match
        self.exclude_packages = coalesce(exclude_packages, [])
        self.include_packages = coalesce(include_packages, [])
        self.exclude_package_files = coalesce(exclude_package_files, [])
        self.include_package_files = coalesce(include_package_files, [])

    @staticmethod
    def matches_includes(name, patterns):
        return all(fnmatch.fnmatchcase(name, pat) for pat in patterns)

    @staticmethod
    def matches_excludes(name, patterns):
        return not any(fnmatch.fnmatchcase(name, pat) for pat in patterns)

    @abc.abstractmethod
    def list_all_packages(self):
        pass

    def list_packages(self):
        all_results = self.list_all_packages()
        ret = PackageListResult(all_results.ok)
        if not ret.ok:
            return ret
        else:
            for result in all_results.results:
                if self.matches_includes(result.name, self.include_packages):
                    if self.matches_excludes(result.name, self.exclude_packages):
                        ret.add_result(name = result.name, link = result.link)
        return ret

    def list_package_files(self, package_name):
        all_results = self.list_all_package_files(package_name)
        ret = PackageFileListResult(all_results.ok)
        if not ret.ok:
            return ret
        else:
            for result in all_results.results:
                if self.matches_includes(result.name, self.include_package_files):
                    if self.matches_excludes(result.name, self.exclude_package_files):
                        ret.add_result(name = result.name, link = result.link, requires_python = result.requires_python)
        return ret
        

    @abc.abstractmethod
    def list_all_package_files(self, package_name):
        pass

    @abc.abstractmethod
    def get_package_file(self, package_name, file_name):
        pass
